
from threading import Event
from uuid import uuid4
import time
import queue

from test import broker_connection_parameter, celery_app
from test.rpc_server import defeat_rpc_processor_name

from kombunicator import RPCServer, RPCClient


received_messages = queue.Queue()


def defeat_answer_processor(data):
    received_messages.put(data)
    print(f"\n\nanswer_processor {data.get('index')}, {data.get('processed')}, {received_messages.qsize()}")


class Test_DefeatableRPCServer:
    listening_key = 'defeat_rpc_key'
    received_messages = list()

    def setup_server(self):
        self.server_ready = Event()
        self.server = RPCServer(
            connection_parameter=broker_connection_parameter,
            listening_key=self.listening_key,
            celery_app=celery_app,
            processing_task_name=defeat_rpc_processor_name,
            ready=self.server_ready
        )

    def start_server(self):
        self.server.start()
        self.server_ready.wait()
        self.server_ready.clear()

    def stop_server(self):
        is_stopped = Event()
        self.server.stop(ready=is_stopped)
        is_stopped.wait()
        self.server.join()

    def get_rpc_client(self):
        client = RPCClient(
            connection_parameter=broker_connection_parameter,
            request_key=self.listening_key,
            default_callback=defeat_answer_processor
        )
        return client

    def test_defeatable_server(self):
        self.setup_server()
        self.start_server()

        client = self.get_rpc_client()

        # put messages while server is running
        for i in range(5):
            data = dict(index=i, unique=uuid4().hex)
            client.request(data)

        # give time for the messages to arrive
        time.sleep(2)
        assert received_messages.qsize() == 5

        # turn off server, put another 5 messages.
        # length of list must remain 5
        self.stop_server()
        for j in range(5):
            data = dict(index=i + j + 1, unique=uuid4().hex)
            client.request(data)

        # give time for the messages to arrive
        time.sleep(2)
        assert received_messages.qsize() == 5

        # turn on server again, put no new messages
        self.setup_server()
        self.start_server()
        # give time for the messages to arrive
        time.sleep(2)
        assert received_messages.qsize() == 10



import json
import uuid
import time
from datetime import datetime
from queue import Queue
from threading import Event
from types import FunctionType, MethodType
from typing import Union

from kombu import Connection, Producer
from strongtyping.strong_typing import match_typing

from kombunicator.MessageConsumer import MessageConsumer
from kombunicator.BusPlugins import RPCRequestPlugin, CallbackPlugin


class RPCClient:

    @match_typing
    def __init__(self, connection_parameter: dict, request_key: str,
                 default_callback: Union[FunctionType, MethodType] = None, plugin: str = None):
        """Create a RPC Client to make request to another service. The message is routed via the RabbitMQ default exchange.

        Parameters
        ----------
        connection_parameter : `dict`
            Holds the connection parameters to a RabbitMQ instance.
            - 'hostname': `str` Host where RabbitMQ is accessible
            - 'virtual_host': `str` virtual host of RabbitMQ instance
            - 'port': `int` port to RabbitMQ instance
            - 'userid': `str` username
            - 'password': `str` password
        request_key: `str`
            The request key as given by the RPC Server.
        default_callback: `callable`
            a default callback function for this RCPClient. Will be called if no function is specified during request.
        plugin: `str`
            Plugin for modifying the request and the callback based on an specific receiver message bus.

        """
        self.connection_parameter = connection_parameter
        self.request_key = request_key
        self.default_callback = default_callback

        self.plugin = plugin

    @property
    def _unique_string(self):
        return str(uuid.uuid4().hex)

    @match_typing(excep_raise=TypeError)
    def request(self, message: dict, headers=None, **kwargs):
        """Requests data from a remote PRC server.

        Parameters
        ----------
        message : dict
            representation of data to be processed remotely
        headers: dict
            additional, optional headers set to the message
        kwargs : dict
            - Can have a 'callback' key with a callable attached.
              Then, the callable will be used as callback to process the
              answer received from the remote RPC server. Otherwise
              the default_callback defined in the __init__() method will
              be used.
            - all other KV pairs of kwargs will be passed
              to the callback as parameters.

        Raises
        ------
            ValueError
        """
        reply_callback = kwargs.pop('callback', self.default_callback)
        self.__request(message=message, headers=headers, reply_callback=reply_callback, **kwargs)

    def sync_request(self, message: dict, headers: json = None,
                     reply_callback: Union[FunctionType, MethodType] = None,
                     timeout: Union[int, float] = 5000, **kwargs):

        reply_callback = reply_callback if reply_callback is not None else self.default_callback
        return self.__request(message=message, headers=headers, reply_callback=reply_callback,
                              timeout=timeout, sync=True, **kwargs)

    @match_typing(excep_raise=TypeError)
    def __request(self, *, message: dict, headers: json = None,
                  reply_callback: Union[FunctionType, MethodType] = None,
                  timeout: Union[int, float] = 5000, sync: bool = False, **kwargs):

        message_correlation_id = self._unique_string
        result_queue_name = self._unique_string
        _ready = Event()

        # create a threaded direct message consumer
        self.consumer_thread = MessageConsumer(
            connection_parameter=self.connection_parameter,
            consumer_type='direct',
            binding_keys=[result_queue_name],
            _thread_ready=_ready
        )
        if sync:
            data = Queue()

        @CallbackPlugin(plugin=self.plugin, message_correlation_id=message_correlation_id)
        def _callback(payload, headers=None, properties=None):
            if properties['correlation_id'] == message_correlation_id:
                if sync:
                    data.queue.append((payload, kwargs))
                else:
                    reply_callback(payload, **kwargs)
                self.consumer_thread.stop()
            else:
                # TODO: report bad message routing and/or wrong correlation id
                pass

        self.consumer_thread.register_message_handler(_callback)
        self.consumer_thread.start()
        _ready.wait()
        self._publish_request(message, result_queue_name=result_queue_name,
                              message_correlation_id=message_correlation_id, headers=headers)

        if sync:
            total_time = 0
            sleeping_time = .1
            while True:
                if data.queue:
                    item, kwgs = data.queue.pop()
                    return reply_callback(item, **kwgs)
                elif total_time >= timeout:
                    break
                else:
                    total_time += sleeping_time
                    time.sleep(sleeping_time)
            # to be sure that the consumer_thread is really stopped
            self.consumer_thread.stop()
            raise TimeoutError

    @RPCRequestPlugin()
    def _publish_request(self, message, *, result_queue_name, message_correlation_id, headers=None):
        with Connection(**self.connection_parameter) as connection:
            with Producer(connection) as producer:
                producer.publish(
                    body=message,
                    exchange='',
                    routing_key=self.request_key,
                    reply_to=result_queue_name,
                    correlation_id=message_correlation_id,
                    headers=headers
                )



import pytest

import kombunicator
from kombunicator.BaseConsumer import BaseConsumer
from kombunicator.ConsumerConfigurator import ConsumerConfigurator

from test import get_random_string


def test_base_consumer_not_implemented():
    consumer = BaseConsumer()
    with pytest.raises(NotImplementedError):
        consumer.process_message({}, {}, {})


def test_consumer_configurator_not_implemented():

    with pytest.raises(NotImplementedError):
        ConsumerConfigurator()

    with pytest.raises(NotImplementedError):
        ConsumerConfigurator.message_handler({}, {}, {})


def test_consumer_configurator_exchange_naming(conn_param):

    exchange_name = 'amq.this_should_throw_exception.exchange'
    binding_keys = ['bk_error']

    # try to bind to an amq. exchange directly
    with pytest.raises(ValueError):
        class InvalidExchangeConsumer(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.binding_keys = binding_keys
                self.consumer_type = 'topic'
                self.q_unique = get_random_string()
                self.accept = ['json']

            @classmethod
            def message_handler(cls, payload, headers, properties):
                pass

        kombunicator.register_message_consumer(InvalidExchangeConsumer)

    exchange_name = 'valid_exchange_name'

    # try and unknown consumer type
    with pytest.raises(ValueError):
        class InvalidTypeConsumer(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.binding_keys = binding_keys
                self.consumer_type = 'unknown_consumer_type'
                self.q_unique = get_random_string()
                self.accept = ['json']

            @classmethod
            def message_handler(cls, payload, headers, properties):
                pass

        kombunicator.register_message_consumer(InvalidTypeConsumer)




import json
import pytest
import random

import kombunicator
from kombunicator import ConsumerConfigurator

from test import get_random_string


# run multiple consumer threads
@pytest.mark.timeout(60)
def test_multi_topic_consumer_setup(tmpdir, conn_param, producer):
    accept = ['json']

    number_of_consumers = random.randint(5, 20)
    number_of_error_consumers = random.randint(10, 30)

    # fire up some topic consumers
    for n in range(number_of_consumers):
        exchange_name = f'test_exchange_topic_{n}'
        binding_keys = [f'test_key_{n}']
        f_name = f'{tmpdir}/test_kombunicator_{n}.tmp'

        class TConsumer(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.consumer_type = 'topic'
                self.binding_keys = binding_keys
                self.q_unique = get_random_string()
                self.accept = accept

            @classmethod
            def message_handler(cls, payload, headers, properties):
                fn = payload.get('f_name', f'{tmpdir}/test_kombunicator_default_{n}.tmp')
                payload['processed'] = True
                with open(fn, 'w') as fh:
                    fh.write(json.dumps(payload))

        kombunicator.register_message_consumer(TConsumer)

        # produce messages to consumer
        test_message = dict(message=get_random_string(), f_name=f_name)
        producer.publish(
            message=test_message,
            exchange=exchange_name,
            routing_key=binding_keys[0]
        )

    for n in range(number_of_error_consumers):
        exchange_name = f'test_error_exchange_topic_{n}'
        binding_keys = [f'test_error_key_{n}']

        class TConsumerError(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.consumer_type = 'topic'
                self.binding_keys = binding_keys
                self.q_unique = get_random_string()
                self.accept = accept

            @classmethod
            def message_handler(cls, payload, headers, properties):
                fn = payload.get('f_name', f'{tmpdir}/test_kombunicator_default_{n}.tmp')
                payload['malformed'] = True
                with open(fn, 'w') as fh:
                    # explicitly dump malformed data into result file
                    fh.write(json.dumps({"message": "Wrong Handler"}))

        kombunicator.register_message_consumer(TConsumerError)

    for n in range(number_of_consumers):
        # load consumed content
        with open(f_name, 'r') as fh:
            received_message = json.loads(fh.read())

        assert test_message['message'] == received_message['message']
        assert 'processed' in received_message
        assert 'malformed' not in received_message

    kombunicator.shutdown_consumers()


@pytest.mark.timeout(60)
def test_multi_fanout_consumer_setup(tmpdir, conn_param, producer):
    accept = ['json']

    number_of_consumers = random.randint(5, 20)
    number_of_error_consumers = random.randint(10, 30)

    # fire up some fanout consumers
    for n in range(number_of_consumers):
        exchange_name = f'test_exchange_fanout_{n}'
        binding_keys = [f'test_key_{n}']
        f_name = f'{tmpdir}/test_kombunicator_{n}.tmp'

        class TConsumer(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.consumer_type = 'fanout'
                self.binding_keys = binding_keys
                self.accept = accept

            @classmethod
            def message_handler(cls, payload, headers, properties):
                fn = payload.get('f_name', f'{tmpdir}/test_kombunicator_default_{n}.tmp')
                payload['processed'] = True
                with open(fn, 'w') as fh:
                    fh.write(json.dumps(payload))

        kombunicator.register_message_consumer(TConsumer)

        # produce messages to consumer
        test_message = dict(message=get_random_string(), f_name=f_name)
        producer.publish(
            message=test_message,
            exchange=exchange_name,
            routing_key=binding_keys[0]
        )

    for n in range(number_of_error_consumers):
        exchange_name = f'test_error_exchange_fanout_{n}'
        binding_keys = [f'test_error_key_{n}']

        class TConsumerError(ConsumerConfigurator):
            def configure(self):
                self.connection_parameter = conn_param
                self.exchange_name = exchange_name
                self.consumer_type = 'fanout'
                self.binding_keys = binding_keys
                self.accept = accept

            @classmethod
            def message_handler(cls, payload, headers, properties):
                fn = payload.get('f_name', f'{tmpdir}/test_kombunicator_default_{n}.tmp')
                payload['malformed'] = True
                with open(fn, 'w') as fh:
                    # explicitly dump malformed data into result file
                    fh.write(json.dumps({"message": "Wrong Handler"}))

        kombunicator.register_message_consumer(TConsumerError)

    for n in range(number_of_consumers):
        # load consumed content
        with open(f_name, 'r') as fh:
            received_message = json.loads(fh.read())

        assert test_message['message'] == received_message['message']
        assert 'processed' in received_message
        assert 'malformed' not in received_message

    kombunicator.shutdown_consumers()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 19.06.20
@author: eisenmenger
"""
import pytest

from kombunicator.utils import get_version


@pytest.fixture(scope="module")
def release_file(tmpdir_factory):
    fn = tmpdir_factory.mktemp('info').join('test_release_info')
    with open(fn, 'w') as file:
        file.write('VERSION_RELEASE=4.5.6\n')
        file.write('DEPLOY_VERSION=1.1.1\n')
        file.write('RELEASE_VERSION=2.32.97')
    return fn


dummy_data = [('RELEASE_VERSION', '2.32.97'), ('DEPLOY_VERSION', '1.1.1'), ('VERSION_RELEASE', '4.5.6')]


@pytest.mark.parametrize('version_string, expected', dummy_data)
def test_version_number(release_file, version_string, expected):
    assert get_version(file_name=release_file, version_string=version_string) == expected


def test_existing_release_info():
    assert get_version(file_name='release_info', version_string='RELEASE_VERSION') != '0.0.1'



import json
from unittest import mock

import kombunicator
import pathlib
import pytest
import time

from copy import copy
from kombunicator.MessageProducer import MessageProducer
from test import get_random_string

msg_handler_result = {}


@pytest.mark.timeout(60)
def test_message_producer(tmpdir, base_test_consumer, producer):
    exchange_name = 'kombunicator_single_producer'
    binding_keys = ['bk_test_single_producer']

    # name of the result file
    tmp_file = pathlib.Path(f'{tmpdir}/test_single_consumer.tmp')

    # define a topic consumer
    class TestConsumer(base_test_consumer):

        @classmethod
        def message_handler(cls, payload, headers, properties):
            result = dict()
            result['processed'] = True
            result['payload'] = payload
            result['headers'] = headers
            with tmp_file.open('w') as fh:
                fh.write(json.dumps(result))

    kombunicator.register_message_consumer(TestConsumer)

    # produce messages to consumer
    test_message = dict(message=get_random_string())
    test_headers = dict(data=get_random_string())

    producer.publish(
        message=test_message,
        headers=test_headers,
        exchange=exchange_name,
        routing_key=binding_keys[0]
    )

    # make sure to receive answer messages.
    time.sleep(0.1)

    # load consumed content
    with tmp_file.open('r') as fh:
        received_message = json.loads(fh.read())

    assert test_message['message'] == received_message['payload']['message']
    assert test_headers['data'] == received_message['headers']['data']
    assert 'processed' in received_message

    kombunicator.shutdown_consumers()


@pytest.mark.timeout(60)
def test_message_producer_multi_binding_keys(base_test_consumer, producer):
    exchange_name = 'kombunicator_single_producer'
    sub_binding_keys = ['bk_test.num_5', 'bk_test.num_5.sub_1']
    range_keys = range(1, 5)
    binding_keys = list(f'bk_test_{i}' for i in range_keys)
    binding_keys.extend(sub_binding_keys)

    # define a topic consumer
    class TestConsumer(base_test_consumer):
        def configure(self):
            super().configure()
            self.binding_keys = binding_keys

        @classmethod
        def message_handler(cls, payload, headers, properties):
            result = dict()
            result['processed'] = True
            result['payload'] = payload
            result['headers'] = headers
            msg_handler_result[properties.get('correlation_id', headers['data'])] = result

    kombunicator.register_message_consumer(TestConsumer)
    messages, headers = [], []

    for binding_key in binding_keys:
        # produce messages to consumer
        test_message = {'message': get_random_string()}
        test_headers = {'data': get_random_string()}

        producer.publish(
            message=test_message,
            headers=test_headers,
            exchange=exchange_name,
            routing_key=binding_key,
            correlation_id=f'test_{binding_key}'
        )
        messages.append(test_message)
        headers.append(test_headers)

    # make sure to receive answer messages.
    time.sleep(0.1)
    kombunicator.shutdown_consumers()

    for i in range_keys:
        bk = f'test_bk_test_{i}'
        assert msg_handler_result[bk]['payload'] == messages[i - 1]
        assert msg_handler_result[bk]['headers'] == headers[i - 1]

    for index, key in enumerate(reversed(sub_binding_keys), start=1):
        assert msg_handler_result[f'test_{key}']['payload'] == messages[-index]
        assert msg_handler_result[f'test_{key}']['headers'] == headers[-index]


def test_configurable_retry_policy(conn_param, default_retry_policy):
    msg_prod = MessageProducer(conn_param)
    assert msg_prod._retry_policy == default_retry_policy

    # change only one field
    msg_prod = MessageProducer(conn_param, retry_policy={'max_retries': 100})
    expected_retry_policy = {
        'interval_start': 0,
        'interval_step': 2,
        'interval_max': 30,
        'max_retries': 100
    }
    assert msg_prod._retry_policy == expected_retry_policy

    # change all fields
    expected_retry_policy = {
        'interval_start': 10,
        'interval_step': 1,
        'interval_max': 60,
        'max_retries': 99
    }
    msg_prod = MessageProducer(conn_param, retry_policy=expected_retry_policy)
    assert msg_prod._retry_policy == expected_retry_policy

    # try to include not existing field (not existing in default retry policy)
    msg_prod = MessageProducer(conn_param, retry_policy={'bad_key': 'breaking', 'maxretries': 1})
    assert msg_prod._retry_policy == default_retry_policy


@mock.patch('kombu.Producer.publish')
def test_message_producer_uses_updated_retry_policy(mocked_producer, conn_param, default_retry_policy):
    new_retry_policy = copy(default_retry_policy)
    new_retry_policy['interval_start'] = 21
    new_retry_policy['max_retries'] = 42

    msg_prod = MessageProducer(conn_param, retry_policy=new_retry_policy)

    exchange_name = 'kombunicator_single_producer'
    binding_keys = ['bk_test_single_producer']
    test_message = {'message': get_random_string()}
    test_headers = {'data': get_random_string()}

    to_publish = {
        'message': test_message,
        'headers': test_headers,
        'exchange': exchange_name,
        'routing_key': binding_keys[0],
        'correlation_id': f'test_{binding_keys[0]}'
    }
    msg_prod.publish(**to_publish)

    assert mocked_producer.called
    mocked_producer.assert_called_with(
        body=to_publish['message'],
        headers=to_publish['headers'],
        exchange=to_publish['exchange'],
        routing_key=to_publish['routing_key'],
        correlation_id=to_publish['correlation_id'],
        retry_policy=new_retry_policy

    )




import json
import pytest
import time

import kombunicator
from kombunicator import ConsumerConfigurator

from test import get_random_string


def test_consumer_class_fails_not_implemented():

    class TestConsumer(ConsumerConfigurator):
        pass

    with pytest.raises(NotImplementedError):
        TestConsumer()


def test_consumer_class_fails_not_implemented_message_handler():

    class TestConsumer(ConsumerConfigurator):

        def configure(self):
            self.exchange_name = 'exchange_name'
            self.connection_parameter = {}
            self.binding_keys = ['_key1', '_key2']
            self.consumer_type = 'topic'
            self.q_unique = get_random_string()
            self.accept = ['json']

    with pytest.raises(NotImplementedError):
        TestConsumer()


def test_consumer_class_fails_not_all_members():
    class TestConsumer(ConsumerConfigurator):

        def configure(self):
            """
            exchange_name is missing
            """
            self.connection_parameter = {}
            self.binding_keys = ['_key1', '_key2']
            self.consumer_type = 'topic'
            self.q_unique = get_random_string()
            self.accept = ['json']

        @classmethod
        def message_handler(cls, payload, headers, properties):
            pass

    with pytest.raises(NotImplementedError):
        TestConsumer()


def test_consumer_class_fails_not_correct_type():
    class TestConsumer(ConsumerConfigurator):

        def configure(self):
            """
            consumer_type must be str
            """
            self.exchange_name = 'exchange_name'
            self.connection_parameter = {}
            self.binding_keys = ['_key1', '_key2']
            self.consumer_type = ['topic', 'topic2']
            self.q_unique = get_random_string()
            self.accept = ['json']

        @classmethod
        def message_handler(cls, payload, headers, properties):
            pass

    with pytest.raises(TypeError):
        TestConsumer()


@pytest.mark.timeout(60)
def test_single_topic_consumer(tmpdir, conn_param, producer, capsys):
    exchange_name = 'kombunicator_single_consumer_topic'
    binding_keys = ['bk_test_single_consumer']

    # name of the result file
    f_name = f'{tmpdir}/test_single_topic_consumer.tmp'

    # define a topic consumer
    class TestConsumer(ConsumerConfigurator):
        def configure(self):
            self.connection_parameter = conn_param
            self.exchange_name = exchange_name
            self.binding_keys = binding_keys
            self.consumer_type = 'topic'
            self.q_unique = get_random_string()
            self.accept = ['json']

        @classmethod
        def message_handler(cls, payload, headers, properties):
            payload['processed'] = True
            with open(f_name, 'w') as fh:
                fh.write(json.dumps(payload))

    # simple test for str and repr
    consumer = TestConsumer()
    assert consumer.__str__() == consumer.__repr__()

    # start the consumer
    kombunicator.register_message_consumer(TestConsumer)

    # produce messages to consumer
    test_message = dict(message=get_random_string())
    producer.publish(
        message=test_message,
        exchange=exchange_name,
        routing_key=binding_keys[0]
    )

    # make sure to receive message.
    time.sleep(0.2)

    # load consumed content
    with open(f_name, 'r') as fh:
        received_message = json.loads(fh.read())

    assert test_message['message'] == received_message['message']
    assert 'processed' in received_message

    kombunicator.shutdown_consumers()


@pytest.mark.timeout(60)
def test_single_direct_consumer(tmpdir, conn_param, producer, capsys):
    binding_keys = ['bk_test_single_consumer_direct']

    # name of the result file
    f_name = f'{tmpdir}/test_single_direct_consumer.tmp'

    # define a topic consumer
    class TestConsumer(ConsumerConfigurator):
        def configure(self):
            self.connection_parameter = conn_param
            self.exchange_name = ''
            self.binding_keys = binding_keys
            self.consumer_type = 'direct'
            self.accept = ['json']

        @classmethod
        def message_handler(cls, payload, headers, properties):
            payload['processed'] = True
            with open(f_name, 'w') as fh:
                fh.write(json.dumps(payload))

    # start the consumer
    kombunicator.register_message_consumer(TestConsumer)

    # produce messages to consumer
    test_message = dict(message=get_random_string())
    producer.publish(
        message=test_message,
        exchange='',
        routing_key=binding_keys[0]
    )

    # make sure to receive message.
    time.sleep(0.2)

    # load consumed content
    with open(f_name, 'r') as fh:
        received_message = json.loads(fh.read())

    assert test_message['message'] == received_message['message']
    assert 'processed' in received_message

    kombunicator.shutdown_consumers()


@pytest.mark.timeout(60)
def test_single_fanout_consumer(tmpdir, conn_param, producer, capsys):
    exchange_name = 'kombunicator_single_consumer_fanout'
    binding_keys = ['bk_test_single_consumer']

    # name of the result file
    f_name = f'{tmpdir}/test_single_fanout_consumer.tmp'

    # define a topic consumer
    class TestConsumer(ConsumerConfigurator):
        def configure(self):
            self.connection_parameter = conn_param
            self.exchange_name = exchange_name
            self.binding_keys = binding_keys
            self.consumer_type = 'fanout'
            self.accept = ['json']

        @classmethod
        def message_handler(cls, payload, headers, properties):
            payload['processed'] = True
            with open(f_name, 'w') as fh:
                fh.write(json.dumps(payload))

    # simple test for str and repr
    consumer = TestConsumer()
    assert consumer.__str__() == consumer.__repr__()

    # start the consumer
    kombunicator.register_message_consumer(TestConsumer)

    # produce messages to consumer
    test_message = dict(message=get_random_string())
    producer.publish(
        message=test_message,
        exchange=exchange_name,
        routing_key=binding_keys[0]
    )

    # make sure to receive message.
    time.sleep(0.2)

    # load consumed content
    with open(f_name, 'r') as fh:
        received_message = json.loads(fh.read())

    assert test_message['message'] == received_message['message']
    assert 'processed' in received_message

    kombunicator.shutdown_consumers()




from celery import Celery
from celery import Signature
from strongtyping.strong_typing import match_typing

from kombunicator.MessageConsumer import MessageConsumer

# How it works:
# RPCServer is a MessageConsumer, i.e. a direct consumer, which is a thread itself.
# RPCServer provides a producer function, which must be registered as a celery task.
# The callback method passed into the RPC server must be a celery task, too.
# result_publisher is then chain attached (celery.chain) to the callback method.


class RPCServer(MessageConsumer):

    @match_typing
    def __init__(self, connection_parameter: dict, listening_key: str, celery_app: Celery,
                 processing_task_name: str, ready):
        super().__init__(
            connection_parameter=connection_parameter,
            consumer_type='direct',
            binding_keys=[listening_key],
            is_rpc=True
        )

        self.celery_app = celery_app
        self._thread_ready = ready

        # the name of the registered celery task is the
        # main application to process the data.
        self.processing_task_name = processing_task_name
        self.register_message_handler(self._on_message)

    def _on_message(self, payload, headers, properties):

        # if the incoming request has headers the processing task of the server has to accept these as an argument.
        task_args = (payload, headers) if headers else (payload, )

        # see: https://stackoverflow.com/questions/31239241/chain-two-remote-tasks-in-celery-by-send-task
        self.celery_app.send_task(
            name=self.processing_task_name,
            args=task_args,
            chain=[
                Signature(
                    'kombunicator.tasks._rpc_server_publisher',
                    kwargs={
                        'connection_parameter': self.connection_parameter,
                        'routing_key': properties.get('reply_to', ''),
                        'correlation_id': properties.get('correlation_id', '')
                    }
                )
            ]
        )



from typing import Union

from kombu import Connection
from kombu import Producer
from kombunicator.Exceptions import AMQPConnectionError
from kombunicator.utils import getter_setter
from strongtyping.strong_typing import match_typing
from kombunicator.BusPlugins import PublishPlugin


class MessageProducer:

    @match_typing
    def __init__(self, connection_parameter: dict, plugin: str = None, retry_policy: dict = None):
        """Create a message producer to publish messages
        to a RabbitMQ instance.

        Parameters
        ----------
        connection_parameter:
            Holds the connection parameters to a RabbitMQ instance.
            - 'hostname': `str` Host where RabbitMQ is accessible
            - 'virtual_host': `str` virtual host of RabbitMQ instance
            - 'port': `int` port to RabbitMQ instance
            - 'userid': `str` username
            - 'password': `str` password
        plugin: `str`
            Plugin for modifying the message based on an specific receiver message bus.

        retry_policy:
            To overwrite the default retry policy which is defined in the following way.
            - 'interval_start': 0,
            - 'interval_step': 2,
            - 'interval_max': 30,
            - 'max_retries': 30

        Raises
        ------
        `kombunicator.AMQPConnectionError`
        """
        self.connection_parameter = connection_parameter
        self._retry_policy = retry_policy if retry_policy is not None else {}

        try:
            conn = Connection(**self.connection_parameter)
            conn.connect()
        except (ConnectionRefusedError, AttributeError, TypeError):
            raise AMQPConnectionError('AMQP connection cannot be established.')
        else:
            # when try was successful
            conn.release()

        self.plugin = plugin

    @PublishPlugin()
    @match_typing
    def publish(self, message: Union[dict, str], headers: dict = {}, exchange: str = '',
                routing_key: str = 'default', correlation_id: str = ''):
        """Publish a message to the RabbitMQ instance.

        Parameters
        ----------
        message : `dict` or `str`
            Message to be published
        headers : `dict`
            Mapping of arbitrary headers to pass along
            with the message
        exchange : `str`
            Name of the exchange to publish a message to
        routing_key : `str`
            Queue routing key
        correlation_id : `str`
            correlation ID to match a specific message
        """
        with Connection(**self.connection_parameter) as connection:
            with Producer(connection) as producer:
                producer.publish(
                    body=message,
                    headers=headers,
                    exchange=exchange,
                    routing_key=routing_key,
                    correlation_id=correlation_id,
                    retry_policy=self._retry_policy
                )

    @getter_setter
    @match_typing
    def _retry_policy(self, value: dict = None) -> dict:
        default_retry_policy = {
            'interval_start': 0,
            'interval_step': 2,
            'interval_max': 30,
            'max_retries': 30
        }
        if value is not None:
            new_retry_policy = self.__check_retry_policy_config__(value, default_retry_policy)
            default_retry_policy.update(new_retry_policy)
            self.__retry_policy = default_retry_policy
        return self.__retry_policy

    @staticmethod
    @match_typing
    def __check_retry_policy_config__(new_config_dict: dict, default_retry_pol: dict) -> dict:
        """
        We allow only keys which are existing in our default retry policy dict
        """
        for bad_key in [k for k in new_config_dict.keys() if k not in default_retry_pol.keys()]:
            del new_config_dict[bad_key]
        return new_config_dict



class CallbackPlugin:

    def __init__(self, plugin=None, **kwargs):
        self.kwargs = kwargs
        self.plugin = plugin

    def __call__(self, function):
        def wrapped_f(payload, headers, properties):
            if not self.plugin:
                return function(payload, headers, properties)
            # C# Rebus can not set properties and therefor the correlation_id will also be expected in the headers and set to properties.
            elif self.plugin == "rebus":
                if not properties:
                    properties = {}
                if not headers.get('correlation_id'):
                    raise ValueError(f'For {self.plugin} plugin the correlation_id field is expected to be present inside the header.')
                if headers['correlation_id'] == self.kwargs["message_correlation_id"]:
                    properties['correlation_id'] = headers['correlation_id']
                    return function(payload, headers, properties)
            else:
                raise ValueError(
                    f'{self.plugin} is not a known messaging plugin.')
        return wrapped_f


class RPCRequestPlugin:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, function):
        def wrapped_f(self, *args, **kwargs):
            if not self.plugin:
                return function(self, *args, **kwargs)
            elif self.plugin == 'rebus':
                if not kwargs.get('headers'):
                    kwargs['headers'] = {}
                # C# Rebus can not set or access properties and therefor the correlation_id will also be expected in the headers.
                kwargs['headers']['correlation_id'] = kwargs.get(
                    "message_correlation_id")
                kwargs['headers']['reply_to'] = kwargs.get("result_queue_name")
                # Rebus Message processing requires a rbs2-content-type header.
                kwargs['headers']['rbs2-content-type'] = "application/json;charset=utf-8"
                return function(self, *args, **kwargs)
            else:
                raise ValueError(
                    f'{self.plugin} is not a known messaging plugin.')
        return wrapped_f


class PublishPlugin:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, function):
        def wrapped_f(self, *args, **kwargs):
            if not self.plugin:
                return function(self, *args, **kwargs)
            elif self.plugin == "rebus":
                if not kwargs.get('headers'):
                    kwargs["headers"] = {}
                # Rebus Message processing requires a rbs2-content-type header.
                kwargs['headers']['rbs2-content-type'] = "application/json;charset=utf-8"
                return function(self, *args, **kwargs)
            else:
                raise ValueError(
                    f'{self.plugin} is not a known messaging plugin.')
        return wrapped_f




import uuid

from kombu import Exchange, Queue, binding
from strongtyping.strong_typing import match_typing

from kombunicator.BaseConsumer import BaseConsumer


class FanoutConsumer(BaseConsumer):

    @match_typing
    def __init__(self, connection, exchange_name: str, binding_keys: list, accept: list = ['json']):
        super().__init__()
        self.connection = connection
        self.exchange_name = exchange_name
        self.binding_keys = binding_keys
        self.accept = accept

        self.exchange = Exchange(self.exchange_name, type='fanout')
        self.q_bindings = [binding(self.exchange, routing_key=key) for key in self.binding_keys]

        self.queue = Queue(
            name=uuid.uuid4().hex,
            exchange=self.exchange,
            bindings=self.q_bindings,
            auto_delete=True,
            exclusive=True,
            expires=2 * 30 * 24 * 60 * 60
        )



import sys
import setuptools
from pathlib import Path


# add kombunicator to path so we can import kombunicator.utils:get_version
ROOT = str(Path(__file__).resolve().parent)
if ROOT not in sys.path:
    sys.path.insert(0, ROOT)


from kombunicator.utils import get_version  # noqa: E402


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kombunicator",
    version=get_version(file_name='release_info', version_string='RELEASE_VERSION'),
    author="Stefan Lasse",
    author_email="stefanlasse87+kombunicator@gmail.com",
    description="A threaded RabbitMQ message producer/consumer and RPC client/server.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/mbio/kombunicator",
    download_url="https://gitlab.com/mbio/kombunicator/-/archive/master/kombunicator-master.tar.gz",
    keywords=["AMQP", "RPC", "kombu", "celery"],
    packages=["kombunicator"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "celery>=4.4.2",
        "kombu>=4.6.8",
        "strongtyping>=1.1.17",
    ],
    python_requires='>=3.6',
    include_package_data=True
)



import time
import pytest
import json
from datetime import datetime
from threading import Event

from test import celery_app

from kombunicator import RPCServer


@pytest.mark.timeout(60)
def test_rpc_server_start_stop(conn_param, tmpdir):

    data_processor_name = 'test_processor'

    @celery_app.task(name=data_processor_name)
    def process_request(data):
        data['processed'] = str(datetime.now())
        return data

    test_server_ready = Event()
    test_server = RPCServer(
        connection_parameter=conn_param,
        listening_key='rpc_test_server_key',
        celery_app=celery_app,
        processing_task_name=data_processor_name,
        ready=test_server_ready
    )

    test_server.start()
    test_server_ready.wait()

    # check if server actually works here via publisher

    test_server.stop()

    # test that server is not running anymore


@pytest.mark.timeout(60)
def test_rpc_server_message_processing(tmpdir, conn_param):

    f_name = f'{tmpdir}/rpc_file.json'

    # this task is registered inside the rpc_server file since the celery app is finalized there
    data_processor_write_to_file = 'write_to_file_processor'

    test_server_ready = Event()
    test_server = RPCServer(
        connection_parameter=conn_param,
        listening_key='rpc_test_server_key',
        celery_app=celery_app,
        processing_task_name=data_processor_write_to_file,
        ready=test_server_ready
    )

    payload = {'body_key': 'body_value', 'f_name': f_name}
    headers = {}
    properties = {}
    test_server._on_message(payload, headers, properties)
    time.sleep(2)

    with open(f_name) as fh:
        received_message = json.loads(fh.read())

    assert received_message.get('processed')
    assert received_message.get('body_key') == payload["body_key"]
